type cell_info = {
  horizontal : Common.bucket ;
  vertical : Common.bucket
} ;;

type cell_color = White of char | Black ;;
type crossword = (cell_color * cell_info) array array ;;


(* Dim 1 = # lines. *)
(* Dim 2 = # de columns. *)
(* .(line).(column) *)
(* Black cells in the middle must be of the form black_nfinit. *)
let new_grid root_transition nb_lines nb_columns =
  (* Final bit set + transition toward END. *)
  let black_finit =
    Int32.logor
      (Int32.shift_right_logical (Int32.of_int Common.end_transition) 10)
      (Int32.of_string "0x200") in
  (* Final bit NOT set + transition toward END. *)
  let black_nfinit = Int32.of_int (Common.end_transition lsl 10) in
  (* Just transition toward the dictionary initial state. *)
  let white_init = Int32.of_int (root_transition lsl 10) in
  (* Core of the crosswords grid. *)
  let grid =
    Array.create_matrix
      nb_lines nb_columns
      (White ' ', { horizontal = white_init ; vertical = white_init }) in
  (* Border of the grid. *)
  for c = 0 to nb_columns - 1 do
    (* Upper borders. *)
    grid.(0).(c) <-
      (Black, { horizontal = black_finit ; vertical =  black_finit}) ;
    (* Lower borders. *)
    grid.(nb_lines - 1).(c) <-
      (Black, { horizontal = black_finit ; vertical = black_nfinit })
  done ;
  for l = 0 to nb_lines - 1 do
    (* Left borders. *)
    grid.(l).(0) <-
      (Black, { horizontal = black_finit ; vertical = black_finit }) ;
    (* Right borders. *)
    grid.(l).(nb_columns - 1) <-
      (Black, { horizontal = black_nfinit ; vertical = black_finit })
  done ;
  (* And now the 4 corners. *)
  grid.(0).(0) <-
    (Black, { horizontal = black_finit ; vertical = black_finit }) ;
  grid.(0).(nb_columns - 1) <-
    (Black, { horizontal = black_finit ; vertical = black_finit }) ;
  grid.(nb_lines - 1).(nb_columns - 1) <-
    (Black, { horizontal = black_finit ; vertical = black_finit }) ;
  grid.(nb_lines - 1).(0) <-
    (Black, { horizontal = black_finit ; vertical = black_finit }) ;
  (* And finally the result grid. *)
  grid
;;



let print_grid grid msg =
  Printf.printf "*** %s ***\n" msg ;
  for lin = 0 to Array.length grid - 1 do
    for col = 0 to Array.length grid.(0) - 1 do
      match grid.(lin).(col) with
      | (Black, _) -> Printf.printf "#"
      | (White c, _) -> Printf.printf "%c" c
    done ;
    Printf.printf "\n" ;
  done ;
  flush stdout
;;


let copy_grid grid =
  let nb_lines = Array.length grid in
  let nb_columns = Array.length grid.(0) in
  let grid' = Array.create_matrix nb_lines nb_columns
      (Black,
       { horizontal = Int32.of_int (-1) ; vertical = Int32.of_int (-1) }) in
  for lin = 0 to nb_lines - 1 do
    for col = 0 to nb_columns - 1 do
      let (color, { horizontal = h ; vertical = v }) = grid.(lin).(col) in
      grid'.(lin).(col) <- (color, { horizontal = h ; vertical = v })
    done
  done ;
  grid'
;;


exception Backtrack ;;


let fill_grid dictionary initial_grid =
  let nb_lines = Array.length initial_grid in
  let nb_columns = Array.length initial_grid.(0) in
  let dic_size = dictionary.Common.size in
  let dic_elems = dictionary.Common.elements in

  let rec try_fill grid index =
    (*
      print_grid grid "CURRENT IN try_fill" ;
     *)
    if index >= nb_lines * nb_columns then
      begin
        (*
          Printf.printf "Index gets out of grid (%d)\n" index ;
          flush stdout ;
         *)
        grid
      end
    else
      let lin = index / nb_lines in
      let col = index mod nb_lines in
      (*
        Printf.printf "*** IN try_fill : (lin, col) = (%d, %d)\n" lin col ;
        flush stdout ;
       *)
      match grid.(lin).(col) with
      | (Black, { horizontal = h ; vertical = v }) ->
          if not (Common.is_final h) || not (Common.is_final v) then
            begin
              (*
                Printf.printf "Black cell NOT FINAL (index = %d)\n" index ;
                flush stdout ;
               *)
              raise Backtrack
            end
          else try_fill grid (index + 1)
      | (White c, { horizontal = h ; vertical = v }) ->
          (* If we can't go further in the dictionary, then it's a failure. *)
          let hbase = Common.get_next_base h in
          let vbase = Common.get_next_base v in
          if hbase = Common.end_transition ||
          vbase = Common.end_transition then
            raise Backtrack
          else
            (* Dtermine the possible characters for this cell,
               horizontally and also vertically. *)
            let allowed_chars =
              let tmpl = ref [] in
              for i = 97 to 122 do
                if hbase + i < dic_size && vbase + i < dic_size &&
                  (Common.get_char dic_elems.(hbase + i)) = i &&
                  (Common.get_char dic_elems.(vbase + i)) = i then
                  let next_h = dic_elems.(hbase + i) in
                  let next_v = dic_elems.(vbase + i) in
                  tmpl := (i, (next_h, next_v)) :: ! tmpl
              done ;
              List.rev !tmpl in
            (*
              Printf.printf "$$$ ALLOWED CHARS $$$ : " ;
              List.iter (fun (c, (_, _)) -> Printf.printf "%c" (Char.chr c))
              allowed_chars ;
              flush stdout ;
             *)
            try_chars grid index allowed_chars

  and try_chars grid index = function
    | [] ->
        (*
          Printf.printf "No more possible characters\n" ;
          flush stdout ;
         *)
        raise Backtrack
    | (char, (next_h, next_v)) :: rem ->
        (*
          Printf.printf "!!! CURRENT INDEX IS %d\n" index ;
          print_grid grid "CURRENT IN try_chars" ;
          Printf.printf "Trying character '%c'\n" (Char.chr char) ;
          flush stdout ;
         *)
        let lin = index / nb_lines in
        let col = index mod nb_lines in
        (*
          Printf.printf "*** IN try_chars : (lin, col) = (%d, %d)\n" lin col ;
          flush stdout ;
         *)
        (* We try to generate the grid with this character for this cell. *)
        let  (_, info) = grid.(lin).(col) in
        grid.(lin).(col) <- (White (Char.chr char), info) ;
        (* Update the cell on the right. *)
        let lin' = lin in
        let col' = col + 1 in
        let (right_col, { horizontal = _ ; vertical = right_v }) =
          grid.(lin').(col') in
        grid.(lin').(col') <-
          (right_col, { horizontal = next_h ; vertical = right_v }) ;
        (* Update the cell below. *)
        let lin' = lin + 1 in
        let col' = col in
        let (left_col, { horizontal = left_h ; vertical = _ }) =
          grid.(lin').(col') in
        grid.(lin').(col') <-
          (left_col, { horizontal = left_h ; vertical = next_v }) ;
        try try_fill grid (index + 1) with
        | Backtrack -> try_chars grid index rem
  in try_fill initial_grid 0
;;



let dic_name = ref "" ;;
let nb_lines = ref 2 ;;
let nb_columns = ref 2 ;;

let main () =
  Arg.parse
    [ ("-l", Arg.Int (fun l -> nb_lines := l), "# lines") ;
      ("-c", Arg.Int (fun c -> nb_columns := c), "# columns") ]
    (fun fname -> dic_name := fname)
    ".........." ;
  let dic_file = open_in_bin !dic_name in
  let initial_base = ((input_value dic_file) : Common.base)  in
  let dictionary = ((input_value dic_file) : Common.dictionary) in
  close_in dic_file ;
  let grid = new_grid initial_base !nb_lines !nb_columns in
  print_grid grid "INITIAL" ;
  let grid' = fill_grid dictionary grid in
  print_grid grid' "FINAL"
;;


main () ;;
