let dic_name = ref "" ;;
let hyphenation_flag = ref true ;;
let hyphenation_warning_flag = ref true ;;



(**
   Checks if a word exists in the dictionary (also check in the auxilliary
   dictionary).
   - dictionary : The dictionary in which to perform the search.
   - initial_base : The entry point in the dictionary.
   - word : The word to check existence in the dictionary.
   Returns:
     bool : true if the word is found in the dictionary, false otherwise. *)
let exists dictionary initial_base word =
  Exist.exists dictionary initial_base word ||
  Spellglobal.Auxdic.mem word !Spellglobal.forget_auxdic ||
  Spellglobal.Auxdic.mem word !Spellglobal.private_auxdic
;;



let main () =
  Spellglobal.private_filename :=
    (try
      (Unixutils.gethomedir ()) ^ "/.merimee_private"
    with Unixutils.No_home_dir -> "") ;
  Arg.parse
    [ ("-d", Arg.String (fun s -> dic_name := s),
       "Specify the main dictionary file to use") ;
      ("-o", Arg.String (fun s -> Spellglobal.out_filename := s),
       "Specify where to save the corrected text") ;
      ("-p", Arg.String (fun s -> Spellglobal.private_filename := s),
       "Specify the user's private dictionary to use") ;
      ("-n", Arg.Unit (fun _ -> hyphenation_flag := false),
       "Disable hyphenation heuristic") ;
      ("-q", Arg.Unit (fun _ -> hyphenation_warning_flag := false),
       "Don't warn when a word don't exist but is a combinaison of several roots") ]
    (fun fname -> Spellglobal.in_filename := fname)
    "Spell check and suggest" ;
  (* If the output file name was not specified, we put a default one to prevent
     crashing the old one. *)
  if !Spellglobal.out_filename = "" then
    Spellglobal.out_filename := !Spellglobal.in_filename ^ ".out" ;
  (* Open the dictionary. *)
  let dic_file = open_in_bin !dic_name in
  let initial_base = ((input_value dic_file) : int)  in
  let dictionary = ((input_value dic_file) : Common.dictionary) in
  close_in dic_file ;
  (* Open the file to spell-check. *)
  let in_file = open_in !Spellglobal.in_filename in
  (* Open the output file. *)
  let out_file = open_out !Spellglobal.out_filename in
  (* Processign of the auxilliary dictionary if it exists. *)
  Privatedict.load_private_dictionary () ;
  (* Backup the initial state of the terminal. *)
  let old_term_mode = Unix.tcgetattr Unix.stdin in
  (* Switch the terminal into input without carriage return and without echo.
     Remove spurious characters that may remain in stdin. *)
  Unix.tcflush Unix.stdin Unix.TCIOFLUSH ;
  let new_term_mode = { old_term_mode with
                        Unix.c_icanon = false ;
                        Unix.c_echo = false } in
  Unix.tcsetattr Unix.stdin Unix.TCSANOW new_term_mode ;
  try
    while true do
      let text_line = input_line in_file in
      let lexbuf = Lexing.from_string text_line in
      let go_on = ref true in
      try
        while !go_on do
          begin
            match French.file lexbuf with
            | French.Word (word, position) ->
                begin
                  let uncap_word = String.uncapitalize word in
                  (* Flag telling if "lower-casing" had an effect. *)
                  let need_capitalization = uncap_word.[0] <> word.[0] in
                  try
                    (* The word must be automatically replaced. *)
                    let new_word =
                      Spellglobal.Autoreplace.find uncap_word
                        !Spellglobal.areplace_db in
                    (* Put back an uppercase letter if needed. *)
                    let new_word' =
                      if need_capitalization then String.capitalize new_word
                      else new_word in
                    output_string out_file new_word'
                  with Not_found ->
                    begin
                      (* The word must not be automatically replaced. *)
                      if not (exists dictionary initial_base uncap_word) then
                        begin
                          (* The word doesn't exist, independently of the
                             hyphenation. *)
                          if !hyphenation_flag then
                            begin
                              (* Hyphenation is activated. *)
                              let sub_words =
                                Hyphenation.split_string uncap_word in
                              if List.for_all
                                  (exists dictionary initial_base)
                                  sub_words then
                                begin
                                  (* Indeed the word exists with hyphenation. *)
                                  if !hyphenation_warning_flag then
                                    begin
                                      User.print_warning
                                        text_line position word sub_words ;
                                      let new_word =
                                        User.wait_interaction uncap_word None
                                          need_capitalization in
                                      output_string out_file new_word
                                    end
                                  else   (* We don't emit any warning. *)
                                    output_string out_file word
                                end
                              else
                                begin
                                  (* Even with hyphenation_warning_flag, the
                                     word doesn't exist. *)
                                  Suggest.suggest dictionary initial_base
                                    position uncap_word ;
                                  User.print_header text_line position word ;
                                  let (suggs, count) =
                                    User.print_and_get_suggestions uncap_word
                                      need_capitalization in
                                  let new_word =
                                    User.wait_interaction uncap_word
                                      (Some (suggs, count))
                                      need_capitalization in
                                  output_string out_file new_word
                                end
                            end
                          else
                            begin
                              (*Hyphenation is not activated and the word
                                 doesn't exist. *)
                              Suggest.suggest dictionary initial_base
                                position uncap_word ;
                              User.print_header text_line position word ;
                              let (suggs, count) =
                                User.print_and_get_suggestions uncap_word
                                  need_capitalization in
                              let new_word =
                                User.wait_interaction uncap_word
                                  (Some (suggs, count))
                                  need_capitalization in
                              output_string out_file new_word
                            end
                        end
                      else
                        begin
                          (* The word exists straightline. *)
                          output_string out_file word
                        end
                    end
                end
            | French.Forget garbage -> output_string out_file garbage
          end
        done
      with End_of_file ->
        (* End of line. *)
        go_on := false ;
        (* Add the end of line in the output file. *)
        output_string out_file "\n"
    done
  with End_of_file | User.Quit ->
    (* End of file. *)
    close_in in_file ;
    close_out out_file ;
    (* Restore th initial state of the terminal. *)
    Unix.tcsetattr Unix.stdin Unix.TCSANOW old_term_mode ;
    Privatedict.save_private_dictionary () ;
    Printf.printf "\n" ;
    flush stdout
;;


try main () with Sys_error msg -> Printf.fprintf stderr "Error : %s\n" msg ;;
