exception Quit ;;


let wait_interaction word sugg_info need_capitalization =
  (* If there is no suggestion info, then we come back from a warning. *)
  if sugg_info = None then
    Printf.printf "\n\n[SP] cC)hange <word> f)orget i)nsert qQ)uit\n"
  else
    Printf.printf
      "\n\n[SP] rR)eplace <number> cC)hange <word> f)orget i)nsert qQ)uit\n" ;
  flush stdout ;
  let s = " " in
  let wait = ref true in
  let new_word = ref "" in
  while !wait do
    let _ = Unix.read Unix.stdin s 0 1 in
    match s.[0] with
    | 'f' ->
        (* Always ignore this word for this session. *)
        Spellglobal.forget_auxdic :=
          Spellglobal.Auxdic.add word !Spellglobal.forget_auxdic ;
        (* We do not change this word. *)
        new_word := word ;
        wait := false
    | 'i' ->
        (* Add thsi word into the user's personnal dictionary. *)
        Spellglobal.private_modified := true ;
        Spellglobal.private_auxdic :=
          Spellglobal.Auxdic.add word !Spellglobal.private_auxdic ;
        (* We do not change this word. *)
        new_word := word ;
        wait := false
    | 'Q' | 'q' ->
        (* End of the spell-checking process. *)
        raise Quit
    | 'C' | 'c' as command ->
        (* Change the current word with another word outside the suggestions. *)
        let old_term_mode = Unix.tcgetattr Unix.stdin in
        let new_term_mode = { old_term_mode with
                              Unix.c_icanon = true ;
                              Unix.c_echo = true ;
                              Unix.c_verase = '\127' ;
                              Unix.c_echoe = true } in
        Unix.tcsetattr Unix.stdin Unix.TCSANOW new_term_mode ;
        Printf.printf "?> " ;
        flush stdout ;
        new_word := read_line () ;
        Unix.tcsetattr Unix.stdin Unix.TCSANOW old_term_mode ;
        (match command with
        | 'C' ->
            (* Replacement must be done for the whole text. *)
            Spellglobal.areplace_db :=
              Spellglobal.Autoreplace.add word !new_word
                !Spellglobal.areplace_db ;
        | _ -> ()) ;
        (* We never put an uppercase letter when replacing a word by a word
           outside the suggestions on the occurrence trigerring this "error".
           However, after, the uppercase letter addition process starts acting
           (see in 'spell.ml'). We take the principle that the user wants to
           replace really this occurrence by something "fixed", "hard",
           "textual". *)
        wait := false
    | 'R' | 'r' as command ->
        begin
          match sugg_info with
          | None ->
              (* If there is no suggestion information, then we come back from
                 a warning and we ignore.  *)
              ()
          | Some (suggs, suggs_count) ->
              (* Change the current word by a suggestion. *)
              let old_term_mode = Unix.tcgetattr Unix.stdin in
              let new_term_mode = { old_term_mode with
                                    Unix.c_icanon = true ;
                                    Unix.c_echo = true ;
                                    Unix.c_verase = '\127' ;
                                    Unix.c_echoe = true } in
              Unix.tcsetattr Unix.stdin Unix.TCSANOW new_term_mode ;
              let wait2 = ref true in
              while !wait2 do
                try
                  Printf.printf "?> " ;
                  flush stdout ;
                  let n = read_int () in
                  if n < suggs_count then
                    begin
                      new_word := fst (List.nth suggs n) ;
                      wait2 := false
                    end
                  else
                    begin
                      Printf.printf "Invalid suggestion\n" ;
                      flush stdout
                    end
                with Failure "int_of_string" ->
                  begin
                    Printf.printf "Invalid suggestion\n" ;
                    flush stdout
                  end
              done ;
              Unix.tcsetattr Unix.stdin Unix.TCSANOW old_term_mode ;
              (match command with
              | 'R' ->
                  (* Replacement must be done for the whole text. *)
                  Spellglobal.areplace_db :=
                    Spellglobal.Autoreplace.add word !new_word
                      !Spellglobal.areplace_db ;
              | _ -> ()) ;
              (* Put back an uppercase letter if needed. *)
              if need_capitalization then
                new_word := String.capitalize !new_word ;
              wait := false
        end
    | ' '  ->
        new_word := word ;
        wait := false
    | _ -> ()
  done ;
  !new_word
;;



(* 'word' : raw word, without case change. *)
let print_header text_line word_position word =
  (* Clear screen. Goto 0 0 *)
  Printf.printf "\027[2J\027[0;0H" ;
  (* Print the faulty word. *)
  Printf.printf"\t%s\t\tFile: %s\n\n\n" word !Spellglobal.in_filename ;
  let prefix = String.sub text_line 0 word_position in
  let suffix_start = word_position + (String.length word) in
  let suffix_len = (String.length text_line) - suffix_start in
  let suffix = String.sub text_line suffix_start suffix_len in
  (* Faulty word in inverse video. *)
  Printf.printf "%s\027[7m%s\027[0m%s\n\n" prefix word suffix ;
  flush stdout
;;


let print_warning text_line word_position word roots =
  print_header text_line word_position word ;
  Printf.printf "Found as combinaison of several roots :\n" ;
  (* Print the involved roots. *)
  Printf.printf "\t( " ;
  List.iter (fun r -> Printf.printf "%s " r) roots ;
  Printf.printf ")\n" ;
  flush stdout
;;


let print_and_get_suggestions word need_capitalization =
  (* Print suggestions. *)
  let sugg_lst = Suggest.Suggs.fold
      (fun word weight accu -> (word, weight) :: accu)
      !Suggest.suggestions [] in
  let sorted_sugg_lst =
    Sort.list (fun (wo1, we1) (wo2, we2)-> we1 < we2) sugg_lst in
  let cpt = ref 0 in
  List.iter
    (fun (word, _) ->
      let word' =
        (* Print suggestions with uppercase if needed, but do not change them in
           place. *)
        if need_capitalization then String.capitalize word else word in
      Printf.printf " %2d: %s\n" !cpt word' ; incr cpt)
    sorted_sugg_lst ;
  flush stdout ;
  (sorted_sugg_lst, !cpt)
;;
